#Eneroth Railroad System

#Copyright Julia Christina Eneroth (eneroh3)
module Ene_Railroad

class ToolTrackInsert
  #This tool is used to add one or more new tracks to the model

  #Variables for calculating track
  @@center_distance = 4.5.m
  @@angle = 15 * Math::PI / 180
  @@formation = "straight"
  @@type_of_track = @@type_of_signals = "default"
  
  #Sketchup tool definitions

  def initialize

    #Only keep tracks in selection
    ss = Sketchup.active_model.selection
    ss.each do |i|
      ss.toggle i unless i.attribute_dictionary("ene_railroad")
    end#each

    #Array of 'virtual' tracks to draw outlines of in tool and to insert to model
    #These are just hashes with the track properties, not track objects
    @tracks = []
    #Point and vector from mouse cursor. snaps to existing tracks
    @point = Geom::Point3d.new
    @vector = Geom::Vector3d.new(10.m, 0, 0)

    #Array of what to base @tracks of.
    #These are too hashes and not track objects
    #In calc and copy mode these are the tracks to insert, but transformed so their origin is what's following the mouse cursor
    #In offset and connect mode these are the original tracks to base new tracks on (offset sorts the track in this array though).
    @track_blueprint = []

    #What user is doing
    #Mode is either calc (getting switches etc from center distance & angle), copy, offset or connect
    @mode =  if ss.empty?
          "calc"
        elsif ss.length == 2 && !Track.get_from_group(ss[0]).connections.flatten.include?(Track.get_from_group(ss[1]))
          #There are 2 unconnected tracks selected
          "connect"
        else
          "copy"
        end

    #Index of the end of @track_blueprint is being placed, used for copy and calc mode
    @track_end_index = 0

    #Distance tracks are being offset
    @offset_distance = 0

    #Indexes of ends in selected tracks to connect
    @track_end_indexes = [1, 0]

    #Input point
    @ip = nil

    #Only affects look of tool, not drawn results
    @default_vector_length = 10.m

    #Statusbar texts
    @status_text_place = S.tr "Click to insert track. Tab = Toggle track end, Arrow keys = Along axis."#place is used both when calculating new track(s) and copying
    @status_text_offset = S.tr "Click or press enter to insert tracks."
    @vcb_label_offset = S.tr "Distance"
    @status_text_connect = S.tr "Click anywhere or press enter to insert track. Tab = Toggle ends to connect."

    @cursor = UI.create_cursor(File.join(PLUGIN_ROOT, "ene_railroad", "cursors", "track_insert.png"), 2, 2)

  end#def

  def onSetCursor

    UI.set_cursor @cursor

  end#def

  def enableVCB?
    true #@mode == "offset"#NOTE: only for offset mode
  end

  def activate

    #Initialize pick helper to select track
    @ph = Sketchup.active_model.active_view.pick_helper

    #Initialize input point to move control points with
    @ip = Sketchup::InputPoint.new

    #Prevents select tool from being selected every time this tool is deactivated. This should only happen when dialog is manually closed
    @tool_active = true

    #Open web dialog with tool settings
    self.init_web_dialog

  end#def

  def deactivate(view)

    @tool_active = false
    @dialog.close
    view.invalidate

  end

  def draw(view)

    color = Sketchup::Color.new(162, 162, 255)

    #Path(s)
    view.line_width = 1
    view.drawing_color = Sketchup::Color.new("Black")
    @tracks.each do |i|
      controls = i[:controls]
      path = Track.calc_path controls, i[:curve_algorithm]
      view.draw_polyline path
    end#each

    #Points on track ends
    points = []
    @tracks.each do |i|
      [0, 1].each do |j|
        points << i[:controls][j]
      end
    end
    points.uniq!
    points.delete @point if @mdoe == "calc" || @mode == "copy"
    view.draw_points points, 7, 4, Sketchup::Color.new("Black") unless points.empty?

    if @mode == "calc" || @mode == "copy"
      #Point where track is being placed
      view.draw_points [@point], 10, 2, color

      #Vector line
      view.line_width = 3
      view.drawing_color = color
      view.draw_line @point, @point.offset(@vector)
    end

    #Draw input point
    @ip.draw view

  end#def

  def onReturn(view)

    #Insert tracks
    self.insert_tracks

  end#def

  def onMouseMove(flags, x, y, view)
    #Set control points and vectors from hovered track if any

    @ph.do_pick(x, y)
    picked = @ph.best_picked
    @ip.pick view, x, y

    #Get point and vector from mouse location
    t_h = Track.get_from_group(picked)
    if t_h
      #Align to existing track
      if t_h.controls[0].distance(@ip.position) < t_h.controls[1].distance(@ip.position)
        #Cursor closest to start
        @point = t_h.controls[0]
        @vector = t_h.controls[2].reverse
      else
        #Cursor closest to end
        @point = t_h.controls[1]
        @vector = t_h.controls[3].reverse
      end
    else
      #Start at cursor
      @point = @ip.position
    end
    @vector.length = @default_vector_length

    if @mode == "calc" || @mode == "copy"
      #Placing predefined tracks
      #Transform @track_blueprint into @tracks depending on input point

      #Transform @track_blueprint to cursor location and save to @tracks
      self.update_tracks_place

    elsif @mode == "offset"
      #Offset tracks
      #Get @offset_distance and offset

      #Get end point of selected tracks closest to input point
      #Get distance
      closest_point = nil
      smallest_distance = nil
      end_vector = nil
      @track_blueprint.each do |i|
        [0, 1].each do |j|
          point = i[:controls][j]
          distance = @point.distance point
          next if smallest_distance && distance > smallest_distance
          smallest_distance = distance
          closest_point = point
          end_vector = i[:controls][j+2].clone
          end_vector.reverse! if j == 1
        end#each
      end#each

      #Project input point to line perpendicular to tracks
      line_perp = [closest_point, end_vector.cross(VECTOR_UP)]
      @point = @point.project_to_line line_perp
      @offset_distance = @point.distance closest_point

      #Positive or negative?
      if (MyGeom.flatten_vector(end_vector).cross(closest_point-(@point))).z < 0
        @offset_distance*= -1
      end

      #Write current distance to VCB
      Sketchup.set_status_text(@offset_distance.abs.to_l.to_s, SB_VCB_VALUE)

      self.update_tracks_offset

    end

    view.invalidate

  end#def

  def onLButtonDown(flags, x, y, view)

    #Insert tracks
    self.insert_tracks

  end#def

  def onUserText(text, view)

    if @mode == "offset"
      begin
        distance = text.to_l
      rescue
        UI.messagebox(S.tr("Invalid length."))
        return
      end
      distance*= -1 if @offset_distance < 0  #for the user the direction mouse is moved in is viewed as positive
      @offset_distance = distance
      self.update_tracks_offset
      self.insert_tracks
    end#if

  end#def

  def onKeyDown(key, repeat, flags, view)

    case key
    when 9#tab
      if @mode == "calc" || @mode == "copy"
        #Toggle what end virtual tracks are being 'hold'
        #Only for copy and calc mode

        self.toggle_end_to_place
        view.invalidate

      elsif @mode == "connect"
        #Toggle what track end are being connected

        self.toggle_ends_to_connect
        view.invalidate

      end

      #Set track direction from arrow keys
      #Up = north, down = south, left = west, right = east
    when VK_RIGHT
      @vector = Geom::Vector3d.new(10.m, 0,0) if @mode == "calc" || @mode == "copy"
    when VK_LEFT
      @vector = Geom::Vector3d.new(-10.m, 0,0) if @mode == "calc" || @mode == "copy"
    when VK_UP
      @vector = Geom::Vector3d.new(0, 10.m,0) if @mode == "calc" || @mode == "copy"
    when VK_DOWN
      @vector = Geom::Vector3d.new(0, -10.m,0) if @mode == "calc" || @mode == "copy"
    end

    if [VK_RIGHT, VK_LEFT, VK_UP, VK_DOWN].include?(key) && (@mode == "calc" || @mode == "copy")
      self.update_tracks_place
      view.invalidate
    end

  end

  def resume(view)
    #Reset status text after tool has been temporarily deactivated
    Sketchup.set_status_text(@status_text, SB_PROMPT)
    Sketchup.set_status_text(@vcb_label, SB_VCB_LABEL)
    view.invalidate
  end#def

  #Own definitions
  #Not called from Sketchup itself

  def blueprint_from_selection
    #Loads track data into @blueprints from selected tracks
    #This should always be used when loading existing tracks from selection to make sure the right properties are loeaded.

    ss = Sketchup.active_model.selection
    @track_blueprint = [];
    ss.each do |i|
      t = Track.get_from_group i
      next unless t
      #Clone controls so original track isn't affected by the transformations
      controls = []
      t.controls.each do |i|
        controls << i.clone
      end
      #Hard coded here what properties are copied
      #Not dynamic since so many properties are references to groups or other values that should NOT be copied
      @track_blueprint << { :controls => controls, :curve_algorithm => t.curve_algorithm, :segments => t.segments, :type_of_track => t.type_of_track, :type_of_signals => t.type_of_signals }
    end

  end#def

  def blueprint_ends
    #Returns all lose ends of blueprint tracks as an array where each element is an array containing a point and corresponding vector

    ends = []
    @track_blueprint.each do |i|
      [0, 1].each do |j|

        unuiqe = true
        @track_blueprint.each do |k|
          next if i == k
          [0, 1].each do |l|
            unuiqe = false if i[:controls][j] == k[:controls][l]
          end
        end
        ends << [i[:controls][j], i[:controls][j+2]] if unuiqe

      end
    end

    ends

  end#def

  def init_web_dialog
    #Open web dialog with tool settings

    @dialog = UI::WebDialog.new(S.tr("Add Track"), false, "ene_railroad_insert_track", 400, 200, 500, 100, false)
    @dialog.navigation_buttons_enabled = false
    @dialog.set_background_color @dialog.get_default_dialog_color
    @dialog.set_file(File.join(PLUGIN_ROOT, "ene_railroad", "web_dialogs", "insert_track", "index.html"))
    
    #Translate strings
    js = S.tr_dlg
    
    #Form data
    js << "document.getElementById('mode').value='#{@mode == "calc" ? "copy" : @mode}';"
    js << "document.getElementById('center_distance').value='#{@@center_distance.to_s}';"
    js << "document.getElementById('angle').value='#{(@@angle/Math::PI*180).to_s.gsub(".",",")}';"
    js << "document.getElementById('formation').value='#{@@formation}';"
    
    #Create track type dropdown
    track_type_selector = "<select id=\"type_of_track\" style=\"width: 75px;\">"
    Ene_Railroad.list_installed_track_types.each { |tt| track_type_selector << "<option value=\"#{tt[:id]}\">#{tt[:name]}</option>" }
    track_type_selector << "</select>"
    js << "document.getElementById('track_type_wrapper').innerHTML='#{track_type_selector}';"
    
    js << "document.getElementById('type_of_track').value='#{@@type_of_track}';"
    
    #Create signal type dropdown
    signal_type_selector = "<select id=\"type_of_signals\" style=\"width: 75px;\">"
    Ene_Railroad.list_installed_signal_types.each { |st| signal_type_selector << "<option value=\"#{st[:id]}\">#{st[:name]}</option>" }
    signal_type_selector << "<option value=\"\">#{S.tr "None"}</option>"
    signal_type_selector << "</select>"
    js << "document.getElementById('type_of_signals_wrapper').innerHTML='#{signal_type_selector}';"
    
    js << "document.getElementById('type_of_signals').value='#{@@type_of_signals}';"
    
    js << "document.onkeyup=port_key;"
    js << "init_tabs();"
    #First tab corresponds to calc mode, other modes are selected from second tab
    js << "change_tab(0,#{@mode == "calc" ? 0 : 1});"
    #Enable or disable 'From Selection' depending on existence of selected tracks
    if Sketchup.active_model.selection.any? { |i| i.attribute_dictionary("ene_railroad") }
      js << "document.getElementById('no_selection_notifier').style.display='none';"
    else
      js << "document.getElementById('from_selection').style.display='none';"
    end
    js << "init_radio_buttons();"
    js << "push_changes();"
    js << "form = document.forms[0];"
    js << "form.onclick = push_changes;"
    js << "form.onkeyup = push_changes;"

    #Show dialog
    if WIN
      @dialog.show { @dialog.execute_script js }
    else
      @dialog.show_modal { @dialog.execute_script js }
    end

    #Change style of web dialog into a toolbox (windows only)
    Ene_Railroad.dialog2toolbox
    
    #Form value changes
    @dialog.add_action_callback("push_changes") {

      #Mode is calc if calc tab is opened. Otherwise it's the value of the mode radio buttons
      @mode = @dialog.get_element_value("mode")
      @mode = "calc" if @dialog.get_element_value("calc_tab_opened") == "true"

      #Only keep tracks in selection
      ss = Sketchup.active_model.selection
      ss.each do |i|
        ss.toggle i unless i.attribute_dictionary("ene_railroad")
      end#each

      if @mode == "calc"
        #Calculate tracks from given data
        #Used for switches etc with given parallel distance and angle

        #Set statusbar text
        @status_text = @status_text_place
        @vcb_label = ""
        Sketchup.set_status_text(@status_text, SB_PROMPT)
        Sketchup.set_status_text(@vcb_label, SB_VCB_LABEL)

        #Get values that needs to be validated
        center_distance = @dialog.get_element_value("center_distance")
        angle =  @dialog.get_element_value("angle")

        #Validate values
        begin
          center_distance = center_distance.to_l
        rescue
          UI.messagebox(S.tr("Invalid length."))
          next
        end
        next if center_distance == 0
        next if angle == ""
        if !angle.match(/\A\d+(.|,)*d*\Z/) or !angle.gsub(",", ".").to_i.between?(1, 89)
          UI.messagebox(S.tr("Angle must be between 1 and 89 degrees."))
          next
        end
        angle.gsub! ",", "."
        angle = (angle.to_f) * Math::PI / 180

        #Save input values as class variables (when passed validation)
        @@center_distance = center_distance
        @@angle = angle
        @@formation = @dialog.get_element_value("formation")
        @@type_of_track = @dialog.get_element_value("type_of_track")
        @@type_of_signals = @dialog.get_element_value("type_of_signals")
        @@type_of_signals = nil if @@type_of_signals == ""

        #Reset index of track end being hold
        @track_end_index = 0

        #Calculate tracks
        #tracks should e possible to connect in various ways so their center distance is always the same.
        #This can be seen as a grid (non-perpendicular) where every cell is a parallelogram and every intersection is a slip.
        #The intersections can be replaced by normal turnouts, curves or straights but their location are always the same.
        #the height of the parallelogram is the center distance, d. This means each side is d/sin(a), where a is the angle.

        #Shorter var names for more readable equations
        d = center_distance
        a = angle

        #Calculate points and vectors

        #Points at track ends
        point_start = Geom::Point3d.new
        point_end = Geom::Point3d.new(d/Math.sin(a), 0, 0)

        point_top_left = Geom::Point3d.new(d/(2*Math.sin(a))-d/(2*Math.tan(a)), d/2, 0)
        point_top_right = Geom::Point3d.new(d/(2*Math.sin(a))+d/(2*Math.tan(a)), d/2, 0)
        point_bottom_left = Geom::Point3d.new(d/(2*Math.sin(a))-d/(2*Math.tan(a)), -d/2, 0)
        point_bottom_right = Geom::Point3d.new(d/(2*Math.sin(a))+d/(2*Math.tan(a)), -d/2, 0)

        #Some reoccurring lengths
        
        #On switches there's a short straight track in each end being the switch track.
        #This used to be 1m but to allow for smaller tracks, such as a model, make it center distance divided by 4,5 m
        length_short_track = (d/(4.5.m)).m
        
        #Adapt length of switch tracks for slips to make either an inside or outside slip.
        if @@formation == "double_slip" || @@formation == "single_slip"
          #Assume standard track gauge.
          g = 1.437.m
          length_inner = d/(2*Math.sin(a))-g/(2*Math.tan(a/2))
          length_outer = d/(2*Math.sin(a))-g/(2*Math.tan(a/4))

          if length_inner > 0 || length_outer > 0
            length_short_track = length_outer > length_short_track ? length_outer : length_inner
          end
        end
        length_vector_long = ((point_top_left-(point_bottom_right)).length-2*length_short_track)/3
        length_vector_short = length_short_track/3

        #Vectors at track ends
        vector_start = Geom::Vector3d.new(1,0,0);
        vector_start.length = length_vector_long
        vector_end = vector_start.reverse

        vector_top_left = point_bottom_right - point_top_left
        vector_top_left.length = length_vector_long
        vector_bottom_right = vector_top_left.reverse
        vector_top_right = point_bottom_left - point_top_right
        vector_top_right.length = length_vector_long
        vector_bottom_left = vector_top_right.reverse

        #Short vectors for short tracks
        vector_start_short = vector_start.clone
        vector_start_short.length = length_vector_short
        vector_end_short = vector_start_short.reverse

        vector_top_left_short = vector_top_left.clone
        vector_top_left_short.length = length_vector_short
        vector_bottom_right_short = vector_top_left_short.reverse

        vector_top_right_short = vector_top_right.clone
        vector_top_right_short.length = length_vector_short
        vector_bottom_left_short = vector_top_right_short.reverse

        #Vectors used to find 'inner' points
        vector_offset_start = vector_start.clone
        vector_offset_start.length = length_short_track
        vector_offset_end = vector_end.clone
        vector_offset_end.length = length_short_track

        vector_offset_top_left = vector_top_left.clone
        vector_offset_top_left.length = length_short_track
        vector_offset_top_right = vector_top_right.clone
        vector_offset_top_right.length = length_short_track
        vector_offset_bottom_left = vector_bottom_left.clone
        vector_offset_bottom_left.length = length_short_track
        vector_offset_bottom_right = vector_bottom_right.clone
        vector_offset_bottom_right.length = length_short_track

        #Inner points, used for switches so they start a small distance from the track end so this part can be used as switch track
        point_inner_start = point_start.offset vector_offset_start
        point_inner_end = point_end.offset vector_offset_end

        point_inner_top_left = point_top_left.offset vector_offset_top_left
        point_inner_top_right = point_top_right.offset vector_offset_top_right
        point_inner_bottom_left = point_bottom_left.offset vector_offset_bottom_left
        point_inner_bottom_right = point_bottom_right.offset vector_offset_bottom_right

        #Points and vectors occurring more than once in a track formation MUST be cloned so transformations doesn't change the same point or vector twice
        if @@formation == "turnout_left"

          @track_blueprint = [
            {#switch
              :controls => [point_start, point_inner_start, vector_start_short, vector_end_short],
              :curve_algorithm => "arc",
              :switch_state => [0,1]
            },
            {#straight
              :controls => [point_inner_start.clone, point_end, vector_start, vector_end],
              :curve_algorithm => "arc"
            },
            {#curve
              :controls => [point_inner_start.clone, point_inner_top_right, vector_start.clone, vector_top_right],
              :curve_algorithm => "arc"
            },
            {#short straight after curve
              :controls => [point_inner_top_right.clone, point_top_right, vector_bottom_left_short, vector_top_right_short],
              :curve_algorithm => "arc"
            }
          ]

        elsif @@formation == "turnout_right"

          @track_blueprint = [
            {#switch
              :controls => [point_start, point_inner_start, vector_start_short, vector_end_short],
              :curve_algorithm => "arc"
            },
            {#straight
              :controls => [point_inner_start.clone, point_end, vector_start, vector_end],
              :curve_algorithm => "arc"
            },
            {#curve
              :controls => [point_inner_start.clone, point_inner_bottom_right, vector_start.clone, vector_bottom_right],
              :curve_algorithm => "arc"
            },
            {#short straight after curve
              :controls => [point_inner_bottom_right.clone, point_bottom_right, vector_top_left_short, vector_bottom_right_short],
              :curve_algorithm => "arc"
            }
          ]

        elsif @@formation == "double_slip"

          @track_blueprint = [
            {#switch (start)
              :controls => [point_start, point_inner_start, vector_start_short, vector_end_short],
              :curve_algorithm => "arc",
              :switch_state => [0,1]
            },
            {#switch (top right)
              :controls => [point_inner_top_right, point_top_right, vector_bottom_left_short, vector_top_right_short],
              :curve_algorithm => "arc"
            },
            {#switch (end)
              :controls => [point_end, point_inner_end, vector_end_short.clone, vector_start_short.clone],
              :curve_algorithm => "arc",
              :switch_state => [0,1]
            },
            {#switch (bottom left)
              :controls => [point_inner_bottom_left, point_bottom_left, vector_top_right_short.clone, vector_bottom_left_short.clone],
              :curve_algorithm => "arc"
            },
            {#straight (start - end)
              :controls => [point_inner_start.clone, point_inner_end.clone, vector_start, vector_end],
              :type_of_signals => "arc"
            },
            {#straight (bottom left - top right)
              :controls => [point_inner_bottom_left.clone, point_inner_top_right.clone, vector_bottom_left, vector_top_right],
              :curve_algorithm => "arc"
            },
            {#curve (start - top right)
              :controls => [point_inner_start.clone, point_inner_top_right.clone, vector_start.clone, vector_top_right.clone],
              :curve_algorithm => "arc"
            },
            {#curve (end - bottom left)
              :controls => [point_inner_end.clone, point_inner_bottom_left.clone, vector_end.clone, vector_bottom_left.clone],
              :curve_algorithm => "arc"
            }
          ]

        elsif @@formation == "single_slip"

          @track_blueprint = [
            {#switch (start)
              :controls => [point_start, point_inner_start, vector_start_short, vector_end_short],
              :curve_algorithm => "arc",
              :switch_state => [0,1]
            },
            {#straight (start - end)
              :controls => [point_inner_start.clone, point_end, vector_start, vector_end],
              :curve_algorithm => "arc"
            },
            {#curve
              :controls => [point_inner_start.clone, point_inner_top_right, vector_start.clone, vector_top_right],
              :curve_algorithm => "arc"
            },
            {#switch (top right)
              :controls => [point_inner_top_right.clone, point_top_right, vector_bottom_left_short, vector_top_right_short],
              :curve_algorithm => "arc"
            },
            {#straight (top left - bottom right)
              :controls => [point_inner_top_right.clone, point_bottom_left, vector_top_right.clone, vector_bottom_left],
              :curve_algorithm => "arc"
            }
          ]

        elsif @@formation == "curve"

          @track_blueprint = [
            {
              :controls => [point_start, point_top_right, vector_start, vector_top_right],
              :curve_algorithm => "arc"
            }
          ]

        elsif @@formation == "straight"

          @track_blueprint = [
            {
              :controls => [point_start, point_end, vector_start, vector_end],
              :curve_algorithm => "arc"
            }
          ]

        end
        
        #Set track and signal type of newly calculated tracks
        @track_blueprint.each { |t| t[:type_of_track] = @@type_of_track; t[:type_of_signals] = @@type_of_signals; }
        
        Sketchup.active_model.active_view.invalidate

      elsif @mode == "copy" && !ss.empty?
        #Copy selected tracks in model

        #Set statusbar text
        @status_text = @status_text_place
        @vcb_label = ""
        Sketchup.set_status_text(@status_text, SB_PROMPT)
        Sketchup.set_status_text(@vcb_label, SB_VCB_LABEL)

        #Load selected tracks to @track_blueprint
        self.blueprint_from_selection

        #Hold track copy in its end
        @track_end_index = 0
        self.toggle_end_to_place(false)
        Sketchup.active_model.active_view.invalidate

      elsif @mode == "offset" && !ss.empty?
        #Offset selected tracks in model

        #Set statusbar text
        @status_text = @status_text_offset
        @vcb_label = @vcb_label_offset
        Sketchup.set_status_text(@status_text, SB_PROMPT)
        Sketchup.set_status_text(@vcb_label, SB_VCB_LABEL)
        
        #Load selected tracks to @track_blueprint
        self.blueprint_from_selection
        
        #Order tracks as they are connected
        #Start with first track in selection
        track_blueprint_sorted = [@track_blueprint[0]]

        #Don't try to add the same track to track_blueprints twice
        tracks_used = [@track_blueprint[0]]

        #Double loop so every track comes up when it can be connected
        @track_blueprint.each do
          @track_blueprint.each do |i|
          
            #Don't try to add the same track to track_blueprints twice
            next if tracks_used.include? i
            
            if [i[:controls][0], i[:controls][1]].include?(track_blueprint_sorted[0][:controls][0])
              #tracks is connected to the start of the sorted track array

              controls = i[:controls].dup
              if controls[0] == track_blueprint_sorted[0][:controls][0]
                #Not same direction as those in array, reverse
                controls[0], controls[1] = controls[1], controls[0]
                controls[2], controls[3] = controls[3], controls[2]
                i[:controls] = controls
              end

              track_blueprint_sorted.unshift i
              tracks_used << i

            elsif [i[:controls][0], i[:controls][1]].include?(track_blueprint_sorted[-1][:controls][1])
              #Track is connected to the end of the sorted track array

              controls = i[:controls].dup
              if controls[1] == track_blueprint_sorted[-1][:controls][1]
                #Not same direction as those in array, reverse
                controls[0], controls[1] = controls[1], controls[0]
                controls[2], controls[3] = controls[3], controls[2]
                i[:controls] = controls
              end

              track_blueprint_sorted << i
              tracks_used << i

            end
          end
        end

        #validate results
        #If not every track has made it into track_blueprint_sorted, they aren't selected properly
        if tracks_used.length != @track_blueprint.length
          UI.messagebox(S.tr("All selected tracks must be connected and there cannot be any branches for offset to work."))
          #Revert to copy mode
          @dialog.execute_script "document.getElementById('mode').value='copy';init_radio_buttons();push_changes();"
          next
        end
        
        @track_blueprint = track_blueprint_sorted

        #Enable VCB
        #NOTE: enable vcb here if possible.

      elsif @mode == "connect" && !ss.empty?
        #Copy selected tracks in model

        #Set statusbar text
        @status_text = @status_text_connect
        @vcb_label = ""
        Sketchup.set_status_text(@status_text, SB_PROMPT)
        Sketchup.set_status_text(@vcb_label, SB_VCB_LABEL)

        #Load selected tracks to @track_blueprint
        self.blueprint_from_selection

        #Validate input
        ends = self.blueprint_ends
        if ends.length < 2
          UI.messagebox(S.tr("There must be at least 2 loose ends to connect."))
          #Revert to copy mode
          @dialog.execute_script "document.getElementById('mode').value='copy';init_radio_buttons();push_changes();"
          next
        end

        #Initialize connecting track
        #Find ends closest to each other that doesn't form an existing track
        closest_distance = nil
        closest_ends = []
        0.upto(ends.length-1) do |i|
          i.upto(ends.length-1) do |j|
            next if i == j
            next unless (Track.instances.select { |k| k.uses_point?(ends[i][0]) && k.uses_point?(ends[j][0]) }).empty?#Don't suggest existing track
            distance = ends[i][0].distance(ends[j][0])
            if !closest_distance || distance < closest_distance
              closest_distance = distance
              closest_ends = [i, j]
            end
          end
        end
        @track_end_indexes = closest_ends unless closest_ends.empty?
        self.update_tracks_connect
        Sketchup.active_model.active_view.invalidate

      end
    }

    #Deselct tool when webdialog closes
    @dialog.set_on_close {
      Sketchup.send_action "selectSelectionTool:" if @tool_active
    }

    #Button under selection tab used to switch to select tool
    @dialog.add_action_callback("select_select_tool") {
      Sketchup.send_action "selectSelectionTool:"
    }

    #Open link in default browser
    @dialog.add_action_callback("open_in_default_browser") { |_, callbacks|
      UI.openURL callbacks
    }
    
    #Port key event to tool
    @dialog.add_action_callback("port_key") { |_, callbacks|
      self.onKeyDown callbacks.to_i, false, 0, Sketchup.active_model.active_view#This tool doesn't use flags so it can just be set to 0
    }

  end

  def insert_tracks
    #Inserts @tracks to model
    #Called from mouse click and pressing enter

    #Preload components so loading doesn't break operation
    @tracks.map { |t| t[:type_of_track] } .uniq.each { |tt| Track.preload_track_type(tt) if tt }
    @tracks.map { |t| t[:type_of_signals] } .uniq.each { |st| Track.preload_signal(st) if st }
    
    Sketchup.active_model.start_operation(S.tr("Add Track"), true)

    @tracks.each do |i|

      #Create track
      t = Track.new
      #Loop all properties of track data and add to newly created track
      #(controls, type_of_track etc)
      i.each_pair do |key, value|
        t.instance_variable_set("@" + key.to_s, value)
      end
      
      #Calculate path (not called when assigning controls since instance_variable_set was used)
      t.calc_path!

      #Find connections
      t.update_connections

      #Draw new track
      t.draw_complete

    end#each

    Sketchup.active_model.commit_operation
    Sketchup.active_model.active_view.invalidate

  end#def

  def toggle_end_to_place(increase = true)
    #Toggles end of @track_blueprint to use as its origin.
    #Used when placing tracks(copy or calc mode)
    #increase is set to false when initializing copy since end isn't toggled then, just set to the first

    #List unconnected ends of @track_blueprint
    ends = self.blueprint_ends

    if ends.empty?
      #Closed loop, isn't toggled but just centered

      bb = Geom::BoundingBox.new
      @track_blueprint.each{ |i| bb.add i[:controls][0] and bb.add i[:controls][1] }
      new_point = bb.center
      new_vector = Geom::Vector3d.new(1, 0, 0)

    else
      #Not closed loop, toggle what end is being placed

      if increase
        #Increase index
        @track_end_index += 1
        @track_end_index = 0 if @track_end_index >= ends.length
      end

      new_point = ends[@track_end_index][0]
      new_vector = ends[@track_end_index][1]

    end

    #Transform virtual tracks so the used point is origin
    translation =  Geom::Transformation.translation Geom::Point3d.new - new_point
    rotation = Geom::Transformation.rotation Geom::Point3d.new, VECTOR_UP, MyGeom.angle_in_plane(Geom::Vector3d.new(1,0,0), new_vector)
    transformation = rotation*translation

    @track_blueprint.each do |i|
      i[:controls].each do |j|
        j.transform! transformation
      end
    end

    #Transform @track_blueprint to cursor location and save to @tracks
    self.update_tracks_place

  end

  def toggle_ends_to_connect
    #Toggles one of the 2 ends of @track_blueprint that is about to be connected
    #Called when pressing tab in connect mode

    #Get number of unconnected ends
    ends_num = self.blueprint_ends.length

    #toggle
    @track_end_indexes[0] += 1
    @track_end_indexes[0]+= 1 if @track_end_indexes[0] == @track_end_indexes[1]
    if @track_end_indexes[0] >= ends_num
      @track_end_indexes[0] = 0
      @track_end_indexes[1] += 1
      @track_end_indexes[1] += 1 if @track_end_indexes[0] == @track_end_indexes[1]
      @track_end_indexes[1] = 0 if @track_end_indexes[1] >= ends_num
      @track_end_indexes[0] = 1 if @track_end_indexes[1] == 0
    end

    #Change @tracks so it contains a track connecting these 2 ends
    self.update_tracks_connect

  end

  def update_tracks_place
    #Transforms tracks_place_cursor to cursor location and save it to @tracks
    #Called in mousemove and when track end being placed is toggled

    transformation = Geom::Transformation.axes @point, @vector, VECTOR_UP.cross(@vector), VECTOR_UP
    @tracks = []
    @track_blueprint.each do |i|
      controls = i[:controls]
      controls_transformed = []
      controls.each { |j| controls_transformed << j.transform(transformation) }
      i_transformed = i.dup
      i_transformed[:controls] = controls_transformed
      @tracks << i_transformed
      #Old un-dynamic way to copy data
      #@tracks << {:controls => controls_transformed, :curve_algorithm => i[:curve_algorithm], :type_of_signals => i[:type_of_signals]}
    end#each

  end#def

  def update_tracks_offset
    #Recreate @tracks based on @track_blueprints and @offset_distance
    #call from mousemove and when entering exact length in VCB

    @tracks = []
    @track_blueprint.each do |i|
      controls = []
      i[:controls].each { |j| controls << j.clone}

      new_controls = []

      #Points
      offset_vector_start = MyGeom.flatten_vector(controls[2]).cross(VECTOR_UP)
      offset_vector_end = VECTOR_UP.cross(MyGeom.flatten_vector(controls[3]))
      offset_vector_start.length = offset_vector_end.length = @offset_distance

      new_controls[0] = controls[0].offset(offset_vector_start)
      new_controls[1] = controls[1].offset(offset_vector_end)

      #Vectors (only affect Bezier curves)
      #Create a decent offset approximation
      #NOTE: make better approximation if possible
      vector_between = (controls[0].offset(controls[2]))-(controls[1].offset(controls[3]))#vector between tips of controls vectors when drawn from control points
      offset_plane0 = [controls[0].offset(controls[2]), vector_between.reverse.normalize+(controls[2].normalize)]
      offset_plane1 = [controls[1].offset(controls[3]), vector_between.normalize+(controls[3].normalize)]
      point_control2 = Geom.intersect_line_plane [new_controls[0], controls[2]], offset_plane0
      point_control3 = Geom.intersect_line_plane [new_controls[1], controls[3]], offset_plane1
      new_controls[2] = point_control2 - new_controls[0]
      new_controls[3] = point_control3 - new_controls[1]
      #Vectors must have correct direction and not be to short
      if new_controls[2].length < 1.m || !new_controls[2].samedirection?(controls[2])
        new_controls[2] = controls[2]
        new_controls[2].length = 1.m
      end
      if new_controls[3].length < 1.m || !new_controls[3].samedirection?(controls[3])
        new_controls[3] = controls[3]
        new_controls[3].length = 1.m
      end

      i_transformed = i.dup
      i_transformed[:controls] = new_controls
      @tracks << i_transformed
      #Old un-dynamic way to copy data
      #@tracks << ({:controls => new_controls, :curve_algorithm => curve_algorithm})
    end#each

  end#def

  def update_tracks_connect
    #Update @tracks based on @track_blueprint
    #Called when initializing connect mode or pressing tab

    #List unconnected ends of @track_blueprint
    ends = self.blueprint_ends

    #Get controls
    point_start = ends[@track_end_indexes[0]][0]
    vector_start = ends[@track_end_indexes[0]][1].reverse
    point_end = ends[@track_end_indexes[1]][0]
    vector_end = ends[@track_end_indexes[1]][1].reverse
    
    #Get track and signal type
    track_start = @track_blueprint.find { |t| t[:controls][0..1].include? point_start }
    type_of_track = track_start[:type_of_track]
    type_of_signals = track_start[:type_of_signals]

    #For a somewhat smooth curve if vectors point somewhat towards each others, use half end distance as vector length
    length = (point_start - point_end).length/2
    vector_start.length = length
    vector_end.length = length

    #Set @tracks
    controls = [point_start, point_end, vector_start, vector_end]
    @tracks = [
      {:controls => controls, :curve_algorithm => "c_bezier", :type_of_track => type_of_track, :type_of_signals => type_of_signals}
    ]

  end#def

end#class

end#module
